import asyncio
from asgiref.sync import sync_to_async
from django.utils.decorators import sync_and_async_middleware
from ...bouncer.bounced import Bounced, get_session_app




@sync_and_async_middleware
def BouncedMiddleware(get_response):
    # One-time configuration and initialization goes here.
    if asyncio.iscoroutinefunction(get_response):
        async def middleware(request):
            # Do something here!
            if hasattr(request, "auth"):
                bounced = Bounced.from_auth(request.auth)
                request.bounced = Bounced.from_auth(request.auth)
                setattr(request, "user", bounced.user)
            elif hasattr(request, "session"):
                app = await sync_to_async(get_session_app)()
                request.bounced = Bounced.from_session_app_and_user(app, request.user)
            else:
                request.bounced = None
            response = await get_response(request)
            return response

    else:
        def middleware(request):
            # Do something here!
            if hasattr(request, "auth"):
                bounced = Bounced.from_auth(request.auth)
                request.bounced = bounced
                setattr(request, "user", bounced.user)
            elif hasattr(request, "session"):
                app = get_session_app()
                request.bounced = Bounced.from_session_app_and_user(app, request.user)
            else:
                request.bounced = None
            response = get_response(request)
            return response

    return middleware


import re
import logging
import jwt



logger = logging.getLogger(__name__)

class JwtToken(dict):
    """
    Mimics the structure of `AbstractAccessToken` so you can use standard
    Django Oauth Toolkit permissions like `TokenHasScope`.
    """
    def __init__(self, decoded, user , app, token):
        decoded["scope"] = decoded["scope"] or ""
        self.scopes = decoded["scope"].split(" ")
        self.issuer = decoded["iss"]
        self.roles = decoded["roles"]
        self.app = app
        self.user = user
        self.grant_type = decoded["type"]
        self.token = token
        super(JwtToken, self).__init__(**decoded, token=token, user=user)

    def __getattr__(self, item):
        return self[item]

    def is_valid(self, scopes=None):
        """
        Checks if the access token is valid.
        :param scopes: An iterable containing the scopes to check or None
        """
        return not self.is_expired() and self.allow_scopes(scopes)

    def is_expired(self):
        """
        Check token expiration with timezone awareness
        """
        # Token expiration is already checked
        return False

    def allow_scopes(self, scopes):
        """
        Check if the token allows the provided scopes
        :param scopes: An iterable containing the scopes to check
        """
        if not scopes:
            return True

        provided_scopes = set(self.scopes)
        resource_scopes = set(scopes)

        return resource_scopes.issubset(provided_scopes)






import asyncio
from ..bouncer.bounced import BounceException

import logging
logger = logging.getLogger(__name__)

def bounce(context_or_scope, required_roles=[], required_scopes=[], allowed_types=["m2m","u2m"], anonymous=False, only_jwt=False):
    
    if hasattr(context_or_scope, "bounced"): # is 
        # We are probably dealing with a normal request
        assert context_or_scope.bounced is not None, "No bounced context provided"
        bouncer = context_or_scope.bounced


        bouncer.bounce(required_roles=required_roles, required_scopes=required_scopes, anonymous=anonymous, only_jwt=only_jwt)
        setattr(context_or_scope, "user", bouncer.user)

    elif hasattr(context_or_scope, "_scope"):
        scope = context_or_scope._scope
        assert "bounced" in scope, "Bounced not in context, did you install the middleware?"
        assert scope["bounced"] is not None, "No bounced context provided"
        bouncer = scope["bounced"]

        bouncer.bounce(required_roles=required_roles, required_scopes=required_scopes, anonymous=anonymous, only_jwt=only_jwt)
        setattr(context_or_scope, "user", bouncer.user)


    elif "bounced" in context_or_scope: # We are dealing with a scobe (websocket request)
        bouncer = context_or_scope["bounced"]

        bouncer.bounce(required_roles=required_roles, required_scopes=required_scopes, anonymous=anonymous, only_jwt=only_jwt)
        context_or_scope["user"] = bouncer.user
    else:
        raise Exception("Unknown Request")

    return context_or_scope



def bounced(required_roles=[], required_scopes=[], allowed_types=["m2m","u2m"], anonymous=False, only_jwt=False):


    def real_decorator(function):

        if asyncio.iscoroutinefunction(function):
            async def bounced_function(root, info, *args, **kwargs):
                bounce(info.context, required_roles=required_roles, required_scopes=required_scopes, anonymous=anonymous, only_jwt=only_jwt)
                return await function(root, info, *args, **kwargs)

        else:
            def bounced_function(root, info, *args, **kwargs):
                bounce(info.context, required_roles=required_roles, required_scopes=required_scopes, anonymous=anonymous, only_jwt=only_jwt)
                return function(root, info, *args, **kwargs)


        return bounced_function


    return real_decorator



def bounced_ws(required_roles=[], required_scopes=[], allowed_types=["m2m","u2m"], anonymous=False, only_jwt=False):


    def real_decorator(function):

        if asyncio.iscoroutinefunction(function):
            async def bounced_function(self, *args, **kwargs):
                try:
                    bounce(self.scope, required_roles=required_roles, required_scopes=required_scopes, anonymous=anonymous, only_jwt=only_jwt)
                except BounceException as e:
                    logger.error(f"Closed because of bouncing {e}")
                    await self.close(e)
                
                return await function(self, *args, **kwargs)

        else:
            def bounced_function(self, *args, **kwargs):
                try:
                   bounce(self.scope, required_roles=required_roles, required_scopes=required_scopes, anonymous=anonymous, only_jwt=only_jwt)
                except BounceException as e:
                    logger.error(f"Closed because of bouncing {e}")
                    self.close(e)
                return function(self, *args, **kwargs)


        return bounced_function


    return real_decorator


def bounced_request(required_roles=[], required_scopes=[]):


    def real_decorator(function):

        if asyncio.iscoroutinefunction(function):
            async def bounced_function(request, *args, **kwargs):
                assert hasattr(request, "bounced"), "Bounced not in context, did you install the middleware?"
                assert request.bounced is not None, "No bounced context provided"
                request.bounced.bounce(required_roles=required_roles, required_scopes=required_scopes)
                return await function(request, *args, **kwargs)

        else:
            def bounced_function(request, *args, **kwargs):
                assert hasattr(request, "bounced"), "Bounced not in context, did you install the middleware?"
                assert request.bounced is not None, "No bounced context provided"
                request.bounced.bounce(required_roles=required_roles,required_scopes=required_scopes)
                return function(request, *args, **kwargs)


        return bounced_function


    return real_decorator




import asyncio
from ...middlewares.utils import set_request_async, set_request_sync
from ...utils import check_token_from_request
from django.utils.decorators import sync_and_async_middleware
import logging
from django.core.exceptions import  PermissionDenied
logger = logging.getLogger(__name__)



@sync_and_async_middleware
def JWTTokenMiddleWare(get_response):
    # One-time configuration and initialization goes here.
    if asyncio.iscoroutinefunction(get_response):
        async def middleware(request):
            # Do something here!
            try:
                decoded, token = check_token_from_request(request)
                if decoded:
                    request = await set_request_async(request, decoded, token)
            except Exception as e:
                logger.error(e)
                raise PermissionDenied(str(e))
            response = await get_response(request)
            return response

    else:
        def middleware(request):
            # Do something here!
            try:
                decoded, token = check_token_from_request(request)
                if decoded:
                    request = set_request_sync(request, decoded, token)
            except Exception as e:
                logger.error(e)
                raise PermissionDenied(str(e))
            response = get_response(request)
            return response

    return middleware


from .enums import LokGrantType
from django.db import models# Create your models here.
from django.contrib.auth.models import AbstractUser

class LokUser(AbstractUser):
    """ A reflection on the real User"""
    email = models.EmailField(unique=True)
    roles = models.JSONField(null=True, blank=True)

class LokApp(models.Model):
    client_id = models.CharField(unique=True, max_length=2000)
    name = models.CharField(max_length=2000)
    grant_type = models.CharField(choices=LokGrantType.choices, max_length=2000)


    def __str__(self):
        return f'{self.name}'


from ..models import LokUser, LokApp
from django.core.exceptions import  PermissionDenied
from django.contrib.auth import get_user_model
from django.http.response import HttpResponseBadRequest
from asgiref.sync import async_to_sync, sync_to_async
from ..token import JwtToken

import logging

logger = logging.getLogger(__name__)

import uuid


def update_or_create_herre(decoded):
    if "email" in decoded and decoded["email"] is not None:
        try:
            user = LokUser.objects.get(email=decoded["email"])
            user.roles = decoded["roles"]
            user.save()
        except LokUser.DoesNotExist:
            user = LokUser(email=decoded["email"], username=f"Lok {decoded['email']}")
            user.roles = decoded["roles"]
            user.set_unusable_password()
            user.save()
            logger.warning("Created new user")
    else:
        user = None

    if "client_id" in decoded and decoded["client_id"] is not None:
        try:
            app = LokApp.objects.get(client_id=decoded["client_id"])
        except LokApp.DoesNotExist:
            app = LokApp(client_id=decoded["client_id"], name=decoded["client_app"], grant_type=decoded["type"])
            app.save()
            logger.warning("Created new app")
    else:
        app = None

    return user, app


@sync_to_async
def set_request_async(request, decoded, token):
    user, app = update_or_create_herre(decoded)
    request.auth = JwtToken(decoded, user, app, token)
    request.user = user
    return request

def set_request_sync(request, decoded, token):
    user, app = update_or_create_herre(decoded)
    request.auth = JwtToken(decoded, user, app, token)
    request.user = user
    return request


@sync_to_async
def set_scope_async(scope, decoded, token):
    user, app = update_or_create_herre(decoded)
    scope["auth"] = JwtToken(decoded, user, app, token)
    scope["user"] = user
    return scope



from django.contrib.auth import get_user_model
from django.core.management.base import BaseCommand
from django.conf import settings


class Command(BaseCommand):
    help = "Creates an admin user non-interactively if it doesn't exist"


    def handle(self, *args, **kwargs):
        superusers = settings.SUPERUSERS

        for superuser in superusers:
            User = get_user_model()
            if not User.objects.filter(email=superuser['EMAIL']).exists():
                User.objects.create_superuser(username=superuser['USERNAME'],
                                            email=superuser['EMAIL'],
                                            password=superuser['PASSWORD'])


from ...middlewares.utils import set_scope_async
from ...utils import decode_token
from ...token import JwtToken
from django.utils.decorators import sync_and_async_middleware
import logging
from django.conf import settings
from django.core.exceptions import  PermissionDenied
from django.contrib.auth import get_user_model
import urllib
from asgiref.sync import async_to_sync, sync_to_async
logger = logging.getLogger(__name__)

UserModel = get_user_model()


class JWTChannelMiddleware:
    """
    Custom middleware (insecure) that takes user IDs from the query string.
    """

    def __init__(self, app):
        # Store the ASGI application we were passed
        self.app = app

    async def __call__(self, scope, receive, send):

        # Close old database connections to prevent usage of timed out connections
        # Look up user from query string (you should also do things like
        # check it's a valid user ID, or if scope["user"] is already populated)
        
        qs = urllib.parse.parse_qs(scope["query_string"].decode())
        if "token" in qs:
            try:
                token = qs["token"][0]
                decoded = decode_token(token)
                await set_scope_async(scope, decoded, token)
            except Exception as e:
                logger.error(e)  

        return await self.app(scope, receive, send)


import logging
import jwt
import json
import base64
import re
from django.core.exceptions import ImproperlyConfigured, PermissionDenied
from django.conf import settings


logger = logging.getLogger(__name__)
jwt_re = re.compile(r'Bearer\s(?P<token>[^\s]*)')

def decode_token(token):
    try:
        headers_enc, payload_enc, verify_signature = token.split(".")
    except ValueError:
        raise jwt.InvalidTokenError(f"Token does not conform to standard JWT {token}")

    payload_enc += '=' * (-len(payload_enc) % 4)  # add padding
    payload = json.loads(base64.b64decode(payload_enc).decode("utf-8"))
    algorithms = [settings.LOK.KEY_TYPE]
    public_key = settings.LOK.PUBLIC_KEY
    if not public_key:
        raise ImproperlyConfigured('Missing setting LOK PUBLIC_KEY')
    
    decoded = jwt.decode(token, public_key, algorithms=algorithms)
    return decoded



def token_from_authorization(authorization):
    m = jwt_re.match(authorization)
    if m:
        token = m.group("token")
        return decode_token(token), token
    else:
        logger.error("Not a valid token. Skipping!")
        return False, False


def check_token_from_request(request):

    if request.META.get("HTTP_AUTHORIZATION", "").startswith("Bearer"):
            if not hasattr(request, "user") or request.user.is_anonymous:
                return token_from_authorization(request.META.get("HTTP_AUTHORIZATION"))
    
    return False, False



from django.db.models import TextChoices

class LokGrantType(TextChoices):
    CLIENT_CREDENTIALS = "client-credentials", "Backend (Client Credentials)"
    IMPLICIT = "implicit", "Implicit Grant"
    AUTHORIZATION_CODE = "authorization-code", "Authorization Code"
    PASSWORD = "password", "Password"

    SESSION = "session", "Django Session"



from ..enums import LokGrantType
from ..token import JwtToken
import logging
from django.contrib.sessions.backends.db import SessionStore
import asyncio

class BounceException(Exception):
    pass

logger = logging.getLogger(__name__)

SESSION_APP = None


def get_session_app():
    global SESSION_APP
    if SESSION_APP is None:
        from ..models import LokApp
        SESSION_APP, created = LokApp.objects.get_or_create(client_id="session", name="Session", grant_type=LokGrantType.SESSION.value)
    
    return SESSION_APP


class Bounced:

    def __init__(self, user, app, scopes, token=None, is_jwt=False) -> None:
        self._user = user
        self._app = app
        self._scopes = scopes or []
        self.scopeset = set(self._scopes)
        self.is_jwt = is_jwt
        self.token = token

    @property
    def user(self):
        return self._user

    @property
    def app(self):
        return self._app

    @property
    def scopes(self):
        return self._scopes

    @property
    def roles(self):
        return self.user.roles


    def bounce(self, required_roles=[], required_scopes=[], anonymous=False, only_jwt=False):

        required_scopes = set(required_scopes)
        required_roles = set(required_roles)

        if only_jwt and not self.is_jwt:
            raise BounceException("Only Apps authorized via JWT are allowed here")
       
        if self._app.grant_type == LokGrantType.CLIENT_CREDENTIALS.value:
            if not required_scopes.issubset(self.scopeset):
                raise BounceException(f"App has not the required Scopes. Required {required_scopes}")

        if self._app.grant_type == LokGrantType.IMPLICIT.value:
            if not anonymous and self.user.is_anonymous:
                raise BounceException("Only signed in users are allowed here")
    
        # Scope tests
        if self._app.grant_type == LokGrantType.PASSWORD.value:
            return True


    @classmethod
    def from_auth(cls, auth: JwtToken):
        return cls(auth.user, auth.app, auth.scopes, token=auth.token, is_jwt=True)

    @classmethod
    def from_session_app_and_user(cls, sessionApp, user):
        return cls(user, sessionApp, [])


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['lok',
 'lok.bouncer',
 'lok.management',
 'lok.management.commands',
 'lok.middlewares',
 'lok.middlewares.request',
 'lok.middlewares.scope',
 'lok.migrations']

package_data = \
{'': ['*']}

install_requires = \
['Django>=3.2.7,<4.0.0', 'PyJWT>=2.1.0,<3.0.0']

setup_kwargs = {
    'name': 'lok',
    'version': '0.1.11',
    'description': 'lok is a jwt middleware that works with django channels',
    'long_description': None,
    'author': 'jhnnsrs',
    'author_email': 'jhnnsrs@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)




from django.utils.decorators import sync_and_async_middleware
import logging
from django.conf import settings
from django.core.exceptions import  PermissionDenied
from django.contrib.auth import get_user_model
from ...bouncer.bounced import Bounced, get_session_app
from asgiref.sync import sync_to_async

logger = logging.getLogger(__name__)

UserModel = get_user_model()



class BouncerChannelMiddleware:
    """
    Custom middleware (insecure) that takes user IDs from the query string.
    """

    def __init__(self, app):
        # Store the ASGI application we were passed
        self.app = app

    async def __call__(self, scope, receive, send):

        # Close old database connections to prevent usage of timed out connections
        # Look up user from query string (you should also do things like
        # check it's a valid user ID, or if scope["user"] is already populated)
        if "auth" in scope:
            bounced =  Bounced.from_auth(scope["auth"])
            scope["bounced"]= bounced
            scope["user"] = bounced.user
            scope["app"] = bounced.app
        elif "session" in scope:
            app = await sync_to_async(get_session_app)()
            bounced = Bounced.from_session_app_and_user(app, scope["user"])
            scope["bounced"] = bounced
            scope["app"] = bounced.app
        else:
            scope["bounced"] = None
        return await self.app(scope, receive, send)